/*
 * Common resource accessor classes.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/semaphore.h>
#include <l4/sys/types.h>

#include <stdio.h>

#include <map>
#include <set>

#include <fsserver/access_map.h>
#include <fsserver/flexpage.h>
#include <fsserver/queue.h>

/* Forward declarations. */

class Paging;
class Pages;



/* Filesystem object accessor. */

class Accessor
{
protected:
  /* Queue object coordinating access. */

  QueueObject _queue;

  /* Accounting. */

  unsigned int _attached = 0;

  /* General concurrency guard. */

  l4_cap_idx_t _accessor_semaphore;

  virtual void lock_accessor()
  {
    l4_semaphore_down(_accessor_semaphore, L4_IPC_NEVER);
  }

  virtual void unlock_accessor()
  {
    l4_semaphore_up(_accessor_semaphore);
  }

public:
  explicit Accessor();

  virtual ~Accessor();

  /* Repeatable activation method. */

  virtual void activate()
  {
    lock_accessor();

    if (!_attached)
    {
      _attached++;
      initialise();
    }

    unlock_accessor();
  }

  /* Accounting methods, returning pre-attach, post-detach status. */

  virtual unsigned int attach()
  {
    lock_accessor();

    unsigned int result = _attached++;

    if (!result)
      initialise();

    unlock_accessor();
    return result;
  }

  virtual unsigned int detach()
  {
    lock_accessor();

    if (_attached)
    {
      _attached--;
      if (!_attached)
        finalise();
    }

    unlock_accessor();
    return _attached;
  }

  virtual void initialise() { }

  virtual void finalise() { }

  /* Access coordination. */

  virtual Queue *queue() { return _queue.client(); }
};



/* Filesystem object accessor supporting file operations. */

class FileAccessor : public Accessor
{
protected:
  /* Size of the filesystem object. */

  size_t _size;

public:
  explicit FileAccessor(size_t size=0)
  : Accessor(), _size(size)
  {
  }

  /* Update methods. */

  virtual size_t get_size() { return _size; }

  virtual void set_size(size_t size) { _size = size; }

  /* Support for conventional file operations. */

  virtual long flush() = 0;

  virtual size_t fstat(char *buffer, size_t length) = 0;

  virtual size_t read(size_t start, char *buffer, size_t length) = 0;

  virtual size_t write(size_t start, char *buffer, size_t length) = 0;
};



/* Filesystem object accessor supporting paging. */

class PagingAccessor : public FileAccessor
{
protected:
  /* Page map. */

  AccessMap _map;

  /* Paging coordinator. */

  Paging *_paging;

  /* Flexpage assignment methods. */

  Flexpage *find_flexpage(unsigned long data_position,
                          unsigned long offset, unsigned long flags);

  Flexpage *new_flexpage(unsigned long data_start, unsigned long data_size,
                         unsigned long offset, l4_addr_t hot_spot,
                         unsigned long flags);

  /* Access to the page collection. */

  Pages *pages();

  /* Page data transfer methods. */

  virtual void fill_flexpage_data(Flexpage *flexpage);

  virtual void flush_flexpage_data(Flexpage *flexpage);

public:
  explicit PagingAccessor(Paging *paging, size_t size=0)
  : FileAccessor(size), _paging(paging)
  {
  }

  /* Accounting methods. */

  virtual void finalise();

  /* Accessor configuration. */

  virtual unsigned long get_fileid() = 0;

  /* Flexpage assignment methods. */

  Flexpage *get_flexpage(unsigned long data_start, unsigned long data_size,
                         unsigned long offset, l4_addr_t hot_spot,
                         unsigned long flags);

  /* Flexpage issuing and revocation methods. */

  void queue_flexpage(Flexpage *flexpage);

  void remove_flexpage(Flexpage *flexpage);

  /* Paging methods. */

  virtual void flush_flexpage(Flexpage *flexpage, bool keep=false, bool purge=false);

  virtual void flush_all(unsigned long start, unsigned long end, bool keep=false);
};



/* A paging accessor linked to other accessors. */

class PipePagingAccessor : public PagingAccessor
{
protected:
  PipePagingAccessor *_next;

  /* Page data transfer methods. */

  virtual void fill_flexpage_data(Flexpage *flexpage);

  virtual void flush_flexpage_data(Flexpage *flexpage);

public:
  explicit PipePagingAccessor(Paging *paging)
  : PagingAccessor(paging), _next(0)
  {
  }

  /* Pipe support methods. */

  virtual PipePagingAccessor *new_accessor();

  virtual PipePagingAccessor *next_accessor();

  /* NOTE: These should be eliminated by reorganising the class hierarchy. */

  virtual unsigned long get_fileid() { return 0; }

  virtual long flush() { return -L4_EACCESS; }

  virtual size_t fstat(char *, size_t) { return 0; }

  virtual size_t read(size_t, char *, size_t) { return 0; }

  virtual size_t write(size_t, char *, size_t) { return 0; }
};
