/*
 * A page collection for handling map requests.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/consts.h>
#include <l4/re/env.h>
#include <l4/sys/factory.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pages.h"



/* Initialisation. */

PageQueue::PageQueue()
{ 
  /* Create semaphores. */
  
  _counter_semaphore = l4re_util_cap_alloc();
  _queue_semaphore = l4re_util_cap_alloc();
  
  if (l4_is_valid_cap(_counter_semaphore)) 
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _counter_semaphore);
  
  if (l4_is_valid_cap(_queue_semaphore))
  {
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _queue_semaphore);
    l4_semaphore_up(_queue_semaphore);
  }
}

PageQueue::~PageQueue()
{
  if (l4_is_valid_cap(_counter_semaphore))
    l4re_util_cap_free_um(_counter_semaphore);

  if (l4_is_valid_cap(_queue_semaphore))
    l4re_util_cap_free_um(_queue_semaphore);
}



/* Constructor. */

Pages::Pages(unsigned long limit) : _allocated(0), _limit(limit)
{
  _memory_semaphore = l4re_util_cap_alloc();

  if (l4_is_valid_cap(_memory_semaphore))
  {
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _memory_semaphore);
    l4_semaphore_up(_memory_semaphore);
  }
}



/* Return whether more memory can be allocated. */

bool Pages::can_get_memory()
{
  return (!_limit) || (_allocated < _limit);
}



/* Obtain more memory from the system. */

l4_addr_t Pages::get_memory(unsigned long size)
{
  l4_semaphore_down(_memory_semaphore, L4_IPC_NEVER);

  l4_addr_t block;

  if (!can_get_memory())
    block = 0;
  else if (posix_memalign((void **) &block, L4_PAGESIZE, size))
    block = 0;
  else
    _allocated += size;

  l4_semaphore_up(_memory_semaphore);

  return block;
}



/* Obtain a new flexpage together with a new memory region. A null flexpage is
   returned if no memory could be obtained. */

Flexpage *Pages::new_flexpage()
{
  l4_addr_t address = get_memory(L4_PAGESIZE);

  if (!address)
    return NULL;

  return new Flexpage(address, L4_PAGESIZE);
}



/* Register a flexpage for the given accessor. */

void Pages::queue_flexpage(PagingAccessor *accessor, Flexpage *flexpage)
{
  _queue.push(accessor, flexpage);
}



/* Insert a flexpage without accessor association at the front of the queue. */

void Pages::release_flexpage(Flexpage *flexpage)
{
  _queue.push_front(NULL, flexpage);
}



/* Reserve a flexpage, removing it from the queue temporarily. */

bool Pages::reserve_flexpage(PagingAccessor *accessor, Flexpage *flexpage)
{
  return _queue.remove(accessor, flexpage);
}



/* Recycle the next flexpage from the queue. This obtains a flexpage from the
   queue for exclusive use, then invokes an accessor to retire the flexpage. */

Flexpage *Pages::remove_flexpage()
{
  /* Obtain the first flexpage item from the queue. */

  PagingAccessor *accessor = NULL;
  Flexpage *flexpage = NULL;
  _queue.pop(&accessor, &flexpage);

  /* Remove the flexpage, flushing its contents and unmapping it. */

  if (accessor != NULL)
    accessor->remove_flexpage(flexpage);

  /* Return the flexpage. */

  return flexpage;
}
